# -*- encoding: utf-8 -*-
"""
GObj :class:`GServerSock`
=========================

GObj for manage server socket.

.. autoclass:: GServerSock
    :members:

"""
try:
    import ssl  # Python 2.6+
except ImportError:
    ssl = None

import logging
import errno
import socket
from ginsfsm.c_sock import (
    GSock,
    GSSLSock
)

GSERVERSOCK_FSM = {  # ONLY for documentation. Partial use of GSOCK_FSM.
    'event_list': (
        'EV_CONNECTED: top output',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
        ),
    }
}

GSERVERSOCK_GCONFIG = {  # ONLY for documentation. Partial use of GSOCK_CONFIG.
    'subscriber': [None, None, 0, None,
        "subcriber of all output-events."
        "Default is ``None``, i.e., the parent"
        ],
    'host': [str, '', 0, None, "server host (ip or name)"],
    'port': [int, 0, 0, None, "server port"],
    'ssl_options': [dict, {}, 0, None, "ssl options"],
}


class GServerSock(GSock):
    """  GSock gobj acting as server sock.

    This gobj is derived from :class:`ginsfsm.c_sock.GSock`,
    but it uses the server behaviour of a sock.
    It creates a listening sock at configured (host, port) attributes.

    `GServerSock` can serve SSL traffic with Python 2.6+ and OpenSSL.
    To make this server serve SSL traffic, send the ssl_options dictionary
    argument with the arguments required for the `ssl.wrap_socket` method,
    including "certfile" and "keyfile"::

       ssl_options={
           "certfile": os.path.join(data_dir, "mydomain.crt"),
           "keyfile": os.path.join(data_dir, "mydomain.key"),
       }

    Each incoming connection
    will create a new :class:`ginsfsm.c_sock.GSock` :term:`gobj`,
    that it will be child of the :attr:`subscriber` `gobj`
    (the parent by default).

        .. note:: Once an ``'EV_CONNECTED'`` event has been sent,
           the GServerSock
           doesn't know anything more of the new gsock client.

           The relationship is directly between the
           accepted :class:`ginsfsm.c_sock.GSock` gobj
           and the :attr:`subscriber`.

           The `subcriber` is responsible of receiving all events
           and destroy the new gsock client.


        .. warning::  Remember to destroy the accepted `gobj`
           with :func:`destroy_gobj` when the `gobj` has been disconnected.

           The `subcriber` knows when a new `gobj` has been accepted because it
           receives the ``'EV_CONNECTED'`` event.

           When the `subcriber` receives a ``'EV_DISCONNECTED'`` event must
           destroy the `gobj` because the connection ceases to exist.

    .. ginsfsm::
       :fsm: GSERVERSOCK_FSM
       :gconfig: GSERVERSOCK_GCONFIG

    *Output-Events:*
        * :attr:`'EV_CONNECTED'`: new client socket connected.

          The gobj sending the `'EV_CONNECTED'` event
          is the new client socket gobj.

          Attributes added to the sent :term:`event`:

            * ``peername``: remote address to which the socket is connected.
            * ``sockname``: the socket’s own address.

          The :attr:`subscriber` can know from which server comes the client
          by the name of client gobj.

          If the server gobj has name, the name of client gobj will be:
          'server-name.clisrv-gsock-#'
    """

    def __init__(self):
        GSock.__init__(self)
        self._n_clisrv = 0

    def start_up(self):
        """ Initialization zone.

        Subscribe all events to parent if :attr:`subscriber` is None.

        Start listen to (host,port).
        """
        if self.subscriber is None:
            self.subscriber = self.parent
        self.do_listen()

    def do_listen(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((self.host, self.port))
        logging.info("Listening '%s' host '%s', port %d..." %
                 (self.name, self.host, self.port))
        self.listen(1024)

    def handle_accepted(self, sock, addr):
        # The socket options to set on receiving a connection.
        # It is a list of (level, optname, value) tuples.
        # TCP_NODELAY disables the Nagle algorithm for writes
        # (Waitress already buffers its writes).
        # TODO: check origins for permitted source ip.
        socket_options = [
            (socket.SOL_TCP, socket.TCP_NODELAY, 1),
            ]
        for (level, optname, value) in socket_options:
            sock.setsockopt(level, optname, value)

        # copied from tornado.netutil.py
        if self.ssl_options:
            assert ssl, "Python 2.6+ and OpenSSL required for SSL"
            try:
                sock = ssl.wrap_socket(
                    sock,
                    server_side=True,
                    do_handshake_on_connect=False,
                    **self.ssl_options
                )
            except ssl.SSLError, err:
                if err.args[0] == ssl.SSL_ERROR_EOF:
                    return sock.close()
                else:
                    raise
            except socket.error, err:
                if err.args[0] == errno.ECONNABORTED:
                    return sock.close()
                else:
                    raise

        self._n_clisrv += 1
        if self.name:
            prefix_name = self.name
        else:
            prefix_name = None
        channel = '.clisrv-gsock-%x' % self._n_clisrv

        if self.ssl_options:
            gsock_class = GSSLSock
        else:
            gsock_class = GSock
        clisrv = self.create_gobj(
            prefix_name + channel if prefix_name else None,
            gsock_class,
            self.subscriber,
        )
        clisrv.set_clisrv_socket(sock)
        clisrv.handle_connect()



# -*- encoding: utf-8 -*-
"""
A python library to do development based in Finite State Machines.
"""
__version__ = '0.4.1'
__title__ = 'ginsfsm'
__author__ = 'Ginés Martínez'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2013 Ginés Martínez'


# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# -*- encoding: utf-8 -*-
"""
GObj :class:`GWsgiServer`
=========================

WSGI server.

It uses :class:`ginsfsm.protocols.http.server.c_http_server.GHttpServer`.

.. autoclass:: GWsgiServer
    :members:

"""

from ginsfsm.gobj import GObj
from ginsfsm.protocols.http.server.c_http_server import GHttpServer
from ginsfsm.protocols.wsgi.common.wsgi_response import WsgiResponse
from ginsfsm.protocols.http.server.c_http_server import GHTTPSERVER_GCONFIG


def ac_channel_opened(self, event):
    """ New client opened.
    """


def ac_channel_closed(self, event):
    """ Client closed.
    """


def ac_request(self, event):
    channel = event.source[-1]
    request = event.request
    application = self.select_app(request)
    response = WsgiResponse(request, self, application)
    self.send_event(channel, 'EV_HTTP_RESPONSE', response=response)


GWSGISERVER_FSM = {
    'event_list': (
        'EV_HTTP_CHANNEL_OPENED: bottom input',
        'EV_HTTP_CHANNEL_CLOSED: bottom input',
        'EV_HTTP_REQUEST: bottom input',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_HTTP_CHANNEL_OPENED',  ac_channel_opened,  'ST_IDLE'),
            ('EV_HTTP_CHANNEL_CLOSED',  ac_channel_closed,  'ST_IDLE'),
            ('EV_HTTP_REQUEST',         ac_request,         'ST_IDLE'),
        ),
    }
}

GWSGISERVER_GCONFIG = GHTTPSERVER_GCONFIG.copy()
GWSGISERVER_GCONFIG.update({
    #TODO: implement a multi wsgi application
    'application': [None, None, 0, None, "wsgi application"],
})


class GWsgiServer(GObj):
    """  WSGI Server gobj.

    The incoming connections will create a new :class:`ginsfsm.c_sock.GSock`
    :term:`gobj`,
    that will be child of the :attr:`subscriber` :term:`gobj`.

        .. warning::  Remember destroy the accepted `gobj`
           with :func:`destroy_gobj` when the `gobj` has been disconnected.

           The `subcriber` knows when a new `gobj` has been accepted because it
           receives the ``'EV_CONNECTED'`` event.

           When the `subcriber` receives a ``'EV_DISCONNECTED'`` event must
           destroy the `gobj` because the connection ceases to exist.

    .. ginsfsm::
       :fsm: GWSGISERVER_FSM
       :gconfig: GWSGISERVER_GCONFIG

    *Input-Events:*

        The relationship is directly between the
        accepted :class:`ginsfsm.c_sock.GSock` gobj and the :attr:`subscriber`.

        See :class:`ginsfsm.c_sock.GSock` `input-events`.

    *Output-Events:*

        The relationship is directly between the
        accepted :class:`ginsfsm.c_sock.GSock` gobj and the :attr:`subscriber`.

        See :class:`ginsfsm.c_sock.GSock` `output-events`.
    """

    def __init__(self):
        GObj.__init__(self, GWSGISERVER_FSM, GWSGISERVER_GCONFIG)
        self._n_connected_clisrv = 0

    def start_up(self):
        if self.name and len(self.name):
            prefix_name = self.name
        else:
            prefix_name = None

        self.ghttpserver = self.create_gobj(
            prefix_name + '.sock-server' if prefix_name else None,
            GHttpServer,
            self,
            subscriber=self,
            host=self.host,
            port=self.port,
            origins=self.origins,
            inactivity_timeout=10,
            responseless_timeout=2,
            maximum_simultaneous_requests=0,
        )
        self.serversock = self.ghttpserver.gserversock.socket

        # Used in environ
        self.effective_host, self.effective_port = self.getsockname()
        self.server_name = self._get_server_name(self.host)

    def select_app(self, request):
        # TODO: do a multi-app, application based on url?
        # con sistema de registro de aplicacion/url
        # una wsgi-app podria ser la puerta de injección de eventos
        # sería multi-wsgi-app en paralelo, no en pila como con los wsgi middleware
        return self.application

    def _get_server_name(self, ip):
        """Given an IP or hostname, try to determine the server name."""
        if ip:
            srv_name = str(ip)
        else:
            srv_name = str(self.serversock.socketmod.gethostname())
        # Convert to a host name if necessary.
        for c in srv_name:
            if c != '.' and not c.isdigit():
                return srv_name
        try:
            if srv_name == '0.0.0.0':
                return 'localhost'
            srv_name = self.serversock.socketmod.gethostbyaddr(srv_name)[0]
        except:
            pass
        return srv_name

    def getsockname(self):
        return self.serversock.getsockname()
